from dataclasses import dataclass

import pytest

from .issue_title_parser import ParsedTitle, parse_issue_title


@dataclass(frozen=True)
class ParseTestExample:
    input_value: str
    expected: ParsedTitle


@pytest.mark.parametrize(
    ("example"),
    [
        ParseTestExample(
            input_value="test(#77): test_description",
            expected=ParsedTitle(prefix="test", description="test_description"),
        ),
        ParseTestExample(
            input_value="test: test_description",
            expected=ParsedTitle(prefix="test", description="test_description"),
        ),
    ],
)
def test_issue_title_parsing(example: ParseTestExample):
    assert example.expected == parse_issue_title(example.input_value)



from dataclasses import dataclass
from types import TracebackType

import typer

from herbarium.subprocess_utils import shell_output


def has_uncommitted_changes() -> bool:
    """Returns True if there are uncommitted changes in the git repo"""
    try:
        shell_output("git diff --quiet")  # Any changes in the working tree
        shell_output("git diff --cached --quiet")  # Any changes in the index
        return False
    except RuntimeError:
        return True


def commit_unstaged() -> None:
    """Commits any unstaged changes"""
    if not has_uncommitted_changes():
        return

    commit_msg = typer.prompt(
        "You have uncommitted changes. Please enter a commit message, or leave blank to abort.",
        default="Misc.",
    )
    if commit_msg == "":
        raise RuntimeError("Aborting because of uncommitted changes")

    shell_output(f'git commit -m "{commit_msg}"')


@dataclass
class StagingMigrater:
    has_stashed: bool = False

    def __enter__(self):
        if not has_uncommitted_changes():
            return

        shell_output("git stash")
        self.has_stashed = True

    def __exit__(self, exc_type: type, exc_val: Exception, exc_tb: TracebackType) -> None:
        """Applies the latest stash"""
        if not self.has_stashed:
            return

        shell_output("git stash apply")



import re
from dataclasses import dataclass


@dataclass(frozen=True)
class ParsedTitle:
    prefix: str
    description: str


def sanitise_text_for_git(input_string: str) -> str:
    char2replacement = {
        " ": "-",
        ":": "/",
        ",": "",
        "'": "",
        '"': "",
        "(": "",
        ")": "",
        "[": "",
        "": "",
        "`": "",
        ">": "",
        "<": "",
        "=": "",
    }

    for character, replacement in char2replacement.items():
        input_string = input_string.replace(character, replacement)

    return input_string.replace("--", "-")


def parse_issue_title(issue_title: str) -> ParsedTitle:
    # Get all string between start and first ":"
    prefix = re.findall(r"^(.*?)[:\(]", issue_title)[0]
    description = re.findall(r": (.*)$", issue_title)[0]

    return ParsedTitle(
        prefix=sanitise_text_for_git(input_string=prefix),
        description=sanitise_text_for_git(input_string=description),
    )



import json
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Optional, Protocol

from .issue_title_parser import parse_issue_title
from .subprocess_utils import shell_output


@dataclass(frozen=True)
class Issue:
    entity_id: Optional[str]
    prefix: str
    description: str


class IssueService(Protocol):
    def setup(self) -> None:
        """Any setup needed, including installing CLI tools, etc."""
        ...

    def get_latest_issues(self, in_progress_label: str) -> Sequence[Issue]:
        ...

    def get_issues_assigned_to_me(self, in_progress_label: str) -> Sequence[Issue]:
        ...

    def label_issue(self, issue: Issue, label: str) -> None:
        ...


# XXX: Add get_recent_issues(n: int) to issue_service


class GithubIssueService(IssueService):
    def setup(self) -> None:
        pass

    def _values_to_issue(self, values: dict[str, str]) -> Issue:
        parsed_title = parse_issue_title(values["title"])
        return Issue(
            entity_id=str(values["number"]),
            description=parsed_title.description,
            prefix=parsed_title.prefix,
        )

    def get_latest_issues(self, in_progress_label: str) -> Sequence[Issue]:
        latest_issues = shell_output(
            f"gh issue list --limit 10 --json number,title --search 'is:open -label:{in_progress_label}'"
        )

        if latest_issues is None:
            return []

        return self._parse_github_json_str(latest_issues)

    def get_issues_assigned_to_me(self, in_progress_label: str) -> Sequence[Issue]:
        """Get issues assigned to current user on current repo"""
        my_issues_cmd = shell_output(
            f"gh issue list --search 'is:open assignee:@me -label:{in_progress_label}' --json number,title"
        )

        if my_issues_cmd is None:
            return []

        return self._parse_github_json_str(my_issues_cmd)

    def _parse_github_json_str(self, issue_str: str) -> Sequence[Issue]:
        values = json.loads(issue_str)
        parsed_output = [self._values_to_issue(v) for v in values]
        return parsed_output

    def _create_label(self, label: str) -> None:
        shell_output(f"gh label create {label}")

    def _add_label_to_issue(self, issue: Issue, label: str) -> None:
        shell_output(f'gh issue edit "{issue.entity_id}" --add-label "{label}"')

    def label_issue(self, issue: Issue, label: str) -> None:
        if not issue.entity_id:
            return
        try:
            self._add_label_to_issue(issue, label)
        except Exception:
            try:
                self._create_label(label)
                self._add_label_to_issue(issue, label)
            except Exception as e:
                raise RuntimeError(f"Error labeling issue {issue.entity_id} with {label}") from e



from dataclasses import dataclass
from typing import Protocol

from .issue_service import Issue
from .issue_stringifyer import IssueStringifyer
from .subprocess_utils import interactive_cmd


class QueueManipulator(Protocol):
    issue_parser: IssueStringifyer

    def fork(self, issue: Issue):
        ...

    def add(self, issue: Issue):
        ...

    def sync(self):
        ...

    def submit(self, automerge: bool = False):
        ...


@dataclass(frozen=True)
class GraphiteManipulator(QueueManipulator):
    issue_parser: IssueStringifyer

    def fork(self, issue: Issue):
        issue_info = self.issue_parser.get_issue_info(issue)
        interactive_cmd(
            f'gt create {issue_info.branch_title} --all -m "{issue_info.first_commit_str}"'
        )

    def add(self, issue: Issue):
        issue_info = self.issue_parser.get_issue_info(issue)
        interactive_cmd(
            f'gt create {issue_info.branch_title} --all --insert -m "{issue_info.first_commit_str}"'
        )
        interactive_cmd(f'git commit --allow-empty -m "{issue_info.first_commit_str}"')

    def sync(self):
        interactive_cmd("gt sync --force --restack --delete")

    def submit(self, automerge: bool = False):
        command = "gt submit -m --no-edit --publish"
        if automerge:
            command += " --merge-when-ready"
        interactive_cmd(command)



from collections.abc import Sequence
from dataclasses import dataclass
from typing import Protocol, Union

import questionary
import typer
from rich.console import Console

from .issue_service import Issue
from .parse_issue_title import parse_issue_title

console = Console()


class IssuePresenter(Protocol):
    ten_latest_prompt: str
    manual_prompt: str
    refresh_prompt: str

    def select_issue_dialog(self, issues: Sequence[Issue]) -> Union[Issue, str]:
        ...


@dataclass(frozen=True)
class DefaultIssuePresenter(IssuePresenter):
    ten_latest_prompt: str = "Recent"
    manual_prompt: str = "Manual"
    refresh_prompt: str = "Refresh..."

    def _show_entry_dialog(self) -> str:
        return typer.prompt("Title")

    def _show_selection_dialog(self, issues: Sequence[Issue]) -> str:
        issue_titles = [issue.description for issue in issues]
        return questionary.select(
            "What's next? 🚀",
            choices=[
                self.refresh_prompt,
                self.ten_latest_prompt,
                self.manual_prompt,
                *issue_titles,
            ],
            default=issue_titles[0],
        ).ask()

    def select_issue_dialog(self, issues: Sequence[Issue]) -> Union[Issue, str]:
        selected_issue_title = self._show_selection_dialog(issues=issues)

        if selected_issue_title in [self.refresh_prompt, self.ten_latest_prompt]:
            return selected_issue_title

        if selected_issue_title == self.manual_prompt:
            selected_issue_title = self._show_entry_dialog()
            parsed_title = parse_issue_title(selected_issue_title)
            return Issue(
                entity_id=None, prefix=parsed_title.prefix, description=parsed_title.description
            )

        return next(issue for issue in issues if issue.description == selected_issue_title)



import enum
from collections.abc import Sequence
from types import TracebackType
from typing import Literal, Optional

import typer
from rich import print
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.prompt import Confirm

from .issue_service import Issue
from .registry import issue_services, presenters, queue_manipulators, queue_navigators

app = typer.Typer()

in_progress_label = "in-progress"
issue_service = issue_services["Github"]()
issue_presenter = presenters["Default"]()
queue_navigator = queue_navigators["Graphite"]()
queue_manipulator = queue_manipulators["Graphite"]

from dataclasses import dataclass


@dataclass
class QueueOperation:
    sync_time: Literal["enter", "exit", "none"] = "enter"

    def __enter__(self):
        print(":arrows_clockwise: [bold green]Syncing with remote...[/bold green]")
        if self.sync_time == "enter":
            queue_manipulator.sync()

    def __exit__(self, exc_type: type, exc_val: Exception, exc_tb: TracebackType) -> None:
        if self.sync_time == "exit":
            queue_manipulator.sync()
        queue_navigator.status()


def retry_issue_getting() -> bool:
    retry = Confirm.ask(
        ":palm_tree: No issues assigned to you in this repository. Do you want to retry?",
        default=True,
    )
    return retry


def get_latest_issues() -> Sequence[Issue]:
    with Progress(
        SpinnerColumn(), TextColumn("[progress.description]{task.description}"), transient=True
    ) as progress:
        progress.add_task("Getting latest issues", start=True)
        latest_issues = issue_service.get_latest_issues(in_progress_label=in_progress_label)

    if not latest_issues:
        if retry_issue_getting():
            raise NotImplementedError
        return []

    return latest_issues


def get_my_issues() -> Sequence[Issue]:
    with Progress(
        SpinnerColumn(), TextColumn("[progress.description]{task.description}"), transient=True
    ) as progress:
        progress.add_task("Getting issues assigned to you", start=True)
        my_issues = issue_service.get_issues_assigned_to_me(in_progress_label=in_progress_label)

    if not my_issues:
        if retry_issue_getting():
            raise NotImplementedError
        return []

    return my_issues


def select_issue(issues: Optional[Sequence[Issue]] = None) -> Issue:
    if issues is None:
        issues = get_my_issues()
    selected_issue = issue_presenter.select_issue_dialog(issues)

    if selected_issue is issue_presenter.refresh_prompt:
        return select_issue()
    if selected_issue is issue_presenter.ten_latest_prompt:
        return select_issue(get_latest_issues())
    if isinstance(selected_issue, str):
        raise NotImplementedError

    return selected_issue


class Location(str, enum.Enum):
    front = "front"
    before = "before"
    after = "after"
    back = "back"


class NoOp:
    def __call__(self):
        pass


str2navigation = {
    "front": queue_navigator.go_to_front,
    "before": queue_navigator.move_up_one,
    "after": NoOp(),
    "back": queue_navigator.go_to_back,
}


@app.command()
@app.command(name="a")
@app.command(name="next")
def add(location: Location = Location.after):
    with QueueOperation(sync_time="exit"):
        selected_issue = select_issue()
        str2navigation[location.value]()
        queue_manipulator.add(selected_issue)
        issue_service.label_issue(selected_issue, label=in_progress_label)


@app.command()
@app.command(name="f")
@app.command(name="new")
def fork(location: Location = Location.front):
    with QueueOperation(sync_time="exit"):
        selected_issue = select_issue()
        str2navigation[location.value]()
        queue_manipulator.fork(selected_issue)
        issue_service.label_issue(selected_issue, label=in_progress_label)


@app.command()
def status():
    queue_navigator.status()


@app.command()
def submit(automerge: bool = False):
    with QueueOperation(sync_time="exit"):
        queue_manipulator.submit(automerge=automerge)
        print(":rocket: [bold green]Stack submitted![/bold green]")


if __name__ == "__main__":
    app()



from dataclasses import dataclass
from typing import Protocol

from .issue_service import Issue


@dataclass(frozen=True)
class IssueInfo:
    branch_title: str
    first_commit_str: str


class IssueStringifyer(Protocol):
    def get_issue_info(self, issue: Issue) -> IssueInfo:
        ...


def sanitise_text_for_bash(input_string: str) -> str:
    char_to_remove = ["`"]
    for character in char_to_remove:
        input_string = input_string.replace(character, "")
    return input_string


class DefaultIssueStringifyer(IssueStringifyer):
    def get_issue_info(self, issue: Issue) -> IssueInfo:
        branch_title = self._get_branch_title(issue=issue)
        first_commit_str = self._get_first_commit_str(issue=issue)

        return IssueInfo(branch_title=branch_title, first_commit_str=first_commit_str)

    def _get_branch_title(self, issue: Issue) -> str:
        entity_id_section = "" if issue.entity_id is None else f"/{issue.entity_id}"
        return f"{issue.prefix}{entity_id_section}/{issue.description}"

    def _get_first_commit_str(self, issue: Issue) -> str:
        if issue.entity_id is None:
            first_commit_str = f"{issue.prefix}: {issue.description}"
        else:
            first_commit_str = f"""{issue.prefix}(#{issue.entity_id}): {issue.description}

Fixes #{issue.entity_id}"""

        return sanitise_text_for_bash(first_commit_str)



import pytest

from .issue_service import GithubIssueService


@pytest.mark.skip()
def test_issue_service():
    service = GithubIssueService()
    service.get_issues_assigned_to_me(in_progress_label="in progress")



from herbarium import queue_manipulator

from .issue_presenter import DefaultIssuePresenter, IssuePresenter
from .issue_service import GithubIssueService, IssueService
from .issue_stringifyer import DefaultIssueStringifyer
from .queue_navigator import GraphiteNavigator, QueueNavigator

issue_services: dict[str, type[IssueService]] = {"Github": GithubIssueService}

queue_navigators: dict[str, type[QueueNavigator]] = {"Graphite": GraphiteNavigator}
queue_manipulators: dict[str, queue_manipulator.QueueManipulator] = {
    "Graphite": queue_manipulator.GraphiteManipulator(issue_parser=DefaultIssueStringifyer())
}

presenters: dict[str, type[IssuePresenter]] = {"Default": DefaultIssuePresenter}



from dataclasses import dataclass
from typing import Protocol

from .git_stager import StagingMigrater
from .subprocess_utils import interactive_cmd


class QueueNavigator(Protocol):
    def go_to_front(self):
        ...

    def go_to_back(self):
        ...

    def move_up_one(self):
        ...

    def move_down_one(self):
        ...

    def status(self):
        ...


@dataclass(frozen=True)
class GraphiteNavigator(QueueNavigator):
    def go_to_front(self):
        with StagingMigrater():
            interactive_cmd("gt trunk")
        interactive_cmd("gt up")

    def go_to_back(self):
        with StagingMigrater():
            interactive_cmd("gt top")

    def move_up_one(self):
        interactive_cmd("gt up")

    def move_down_one(self):
        interactive_cmd("gt down")

    def status(self):
        interactive_cmd("gt log short --reverse")
